// ============================================================================
// Project:	A Tiny DBMS (Exercise for Course "DBMS-Implementation")
// Filename:	block.h
// Purpose:	Blocks - Units of e.g. 8K in data files.
// Last Change:	28.01.2012
// Language:	C++
// Author:	Stefan Brass
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2011-2012 by Stefan Brass
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	BLOCK_INCLUDED
#define BLOCK_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef BNO_INCLUDED
#include "bno.h"
#endif

#ifndef BOFF_INCLUDED
#include "boff.h"
#endif

#ifndef BTYPE_INCLUDED
#include "btype.h"
#endif

//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// BLOCK_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long BLOCK_MAGIC = 0x424C4F1AL;	// 'BLO\n'

//-----------------------------------------------------------------------------
// BLOCK_PW_IND: Partial Write Indicator (any bit pattern):
//-----------------------------------------------------------------------------

static const long BLOCK_PW_IND_A = 0x53424442L;	// 'SBDB'
static const long BLOCK_PW_IND_B = 0x73626462L;	// 'sbdb'

//-----------------------------------------------------------------------------
// BLOCK_DATASIZE: Number of data bytes (without header and footer).
//-----------------------------------------------------------------------------

static const unsigned int BLOCK_DATASIZE =
		VER_BLOCKSIZE
		- sizeof(long int) /* Magic Number (even in non-debug mode) */
		- sizeof(btype_t)  /* Block type */
		- sizeof(int)      /* Partial Write Indicator in Header */
		- sizeof(bno_t)    /* Block Number */
		- sizeof(int)      /* Partial Write Indicator in Footer */
		;

//-----------------------------------------------------------------------------
// BLOCK_DATASIZE_INT: Number of ints/unsigned ints in the data part.
//-----------------------------------------------------------------------------

static const unsigned int BLOCK_DATASIZE_INT =
	BLOCK_DATASIZE / sizeof(VER_INT);

//-----------------------------------------------------------------------------
// BLOCK_DATASIZE_SHORT: Number of shorts in the data part of the block.
//-----------------------------------------------------------------------------

static const unsigned int BLOCK_DATASIZE_SHORT =
	BLOCK_DATASIZE / sizeof(VER_USHORT);

//=============================================================================
// Class for Blocks:
//=============================================================================

class block_c {
	public:

//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	// The constructor constructs an empty/untyped block (e.g. for buffer).
	block_c();

	// The destructor marks the block as unusable:
	~block_c();

//-----------------------------------------------------------------------------
// (Object) Methods:
//-----------------------------------------------------------------------------

	// The validity/integrity check is needed also in non-debug mode.
	// Blocks read from the disk must be checked.
	// A null pointer is returned if everything is ok, else an error msg.
	str_t	invalid() const;

	// Query the block type of this block:
	inline btype_t btype() const
		{ CHECK_VALID("block_c::btype");
		  return(BType); }

	// Prepare a block for writing. Change partial write indicators.
	void prepare_write();

	// Set block number:
	inline void set_block_no(bno_t bno)
		{ CHECK_VALID("block_c::set_block_no");
		  BlockNo = bno;
		}

	// Return block number:
	inline bno_t block_no() const
		{ CHECK_VALID("block_c::block_no");
		  return BlockNo;
		}

//-----------------------------------------------------------------------------
// (Object) Methods: *** FOR TESTING ONLY. WILL BE REMOVED IN FUTURE ***
//-----------------------------------------------------------------------------

	// Return byte from data part:
	inline VER_BYTE databyte(boff_t i) const
		{ CHECK_VALID("block_c::databyte");
		  CHECK(i < BLOCK_DATASIZE,
			"block_c::databyte: index too large");
		  return Data.CharData[i];
		}

	// Set byte in data part:
	inline void set_databyte(boff_t i, VER_BYTE b)
		{ CHECK_VALID("block_c::set_databyte");
		  CHECK(i < BLOCK_DATASIZE,
			"block_c::set_databyte: index too large");
		  Data.CharData[i] = b;
		}


//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	block_c(const block_c& block);			// Not implemented
	block_c& operator=(const block_c& block);	// Not implemented

//-----------------------------------------------------------------------------
// Attributes (Protected Members - Accessible From Subclasses):
//-----------------------------------------------------------------------------

	protected:

	// Magic Number (to check object type):
	long int Magic;

	// Block Type:
	btype_t BType;

	// Partial Write Indicator in Header:
	int HeadPartialWriteInd;

	// Block number:
	bno_t BlockNo;

	// Data Part of the Block:
	union {
		VER_BYTE   CharData[BLOCK_DATASIZE];
		VER_USHORT ShortData[BLOCK_DATASIZE_SHORT];
		VER_INT    IntData[BLOCK_DATASIZE_INT];
		VER_UINT   UIntData[BLOCK_DATASIZE_INT];
	} Data;

	// Partial Write Indicator in Footer:
	int FootPartialWriteInd;

//-----------------------------------------------------------------------------
// End of Class:
//-----------------------------------------------------------------------------

};

//-----------------------------------------------------------------------------
// Define pointer type:
//-----------------------------------------------------------------------------

typedef block_c *block_t;

//=============================================================================
// End of Include File:
//=============================================================================

#endif

