// ============================================================================
// Project:	A Tiny DBMS (Exercise for Course "DBMS-Implementation")
// Filename:	check.cpp
// Purpose:	Runtime Assertion Check
// Last Change:	14.12.2011
// Language:	C++
// Author:	Stefan Brass
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 1997-2011 by Stefan Brass
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"

#include <stdlib.h>
#include <iostream>
// #include <iomanip>

#ifdef   VER_SIGNAL
// #include <signal>
#include <signal.h>
#endif

#include "check.h"

//=============================================================================
// Forward Declarations for Local Functions:
//=============================================================================

#ifdef VER_SIGNAL
#ifdef VER_DEBUG
static	void	check_sig(int sig);
#endif
#endif

//=============================================================================
// Global Variables:
//=============================================================================

//-----------------------------------------------------------------------------
// CheckOnExit: Function to be called when check_failed aborts program.
//-----------------------------------------------------------------------------

#ifdef VER_DEBUG
static void (*CheckOnExit)() = 0;
#endif

//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// check_init: Catch "segmentation fault", print error message to stdout.
//-----------------------------------------------------------------------------

#ifdef VER_DEBUG
void	check_init(void)
{
#ifdef VER_SIGNAL
	signal(SIGSEGV, &check_sig);
	signal(SIGBUS, &check_sig);
	signal(SIGFPE, &check_sig);
#endif
}
#endif

//-----------------------------------------------------------------------------
// check_failed: Print error message and abort program.
//-----------------------------------------------------------------------------

#ifdef VER_DEBUG
void	check_failed(str_t file, int line, str_t msg)
{
	// Flush cout:
	std::cout << std::flush;

	// Print error message:
	std::cerr << "\nAssertion failed ("
		<< file << ":" << line << "): "
		<< msg << "\n";

	// Perhaps someday I will automatically send me a mail ...

	// Clean-up Function, e.g. for closing windows:
	if(CheckOnExit)
		(*CheckOnExit)();

	// Terminate program:
#ifdef VER_UNIX
	abort();
#else
	exit(3);
#endif
}
#endif

//-----------------------------------------------------------------------------
// check_emsg: Check whether error message is not nil, print and abort if yes.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
void	check_emsg(str_t file, int line, str_t place, str_t msg)
{
	// If msg is NIL, everything is ok:
	if(!msg)
		return;

	// Else construct error message:
	char buf[128];
	str_cpy(buf, 128, place);
	str_cat(buf, 128, ": ");
	str_cat(buf, 128, msg);

	// Now call "assertion failed" routine:
	check_failed(file, line, buf);
}
#endif

//-----------------------------------------------------------------------------
// check_on_exit: Define a function to be called when check_failed aborts.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
void	check_on_exit(void (*f)())
{
	CheckOnExit = f;
}

#endif


//=============================================================================
// Local Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// check_sig: Signal handler.
//-----------------------------------------------------------------------------

#ifdef VER_SIGNAL
#ifdef VER_DEBUG
void check_sig(int sig)
{
	str_t sig_name;

	// Flush cout:
	std::cout << std::flush;

	// Determine signal name:
	if(sig == SIGSEGV)
		sig_name = "Segmentation Fault";
	else if(sig == SIGBUS)
		sig_name = "Bus Error";
	else if(sig == SIGFPE)
		sig_name = "Arithmetic Error";
	else
		sig_name = "Unexpected Signal";

	// Print error message:
	std::cerr << "Got signal '" << sig_name << "'\n";

	// Perhaps someday I will automatically send me a mail ...

	// Terminate program:
#ifdef VER_UNIX
	abort();
#else
	exit(3);
#endif
}
#endif
#endif

