// ============================================================================
// Project:	A Tiny DBMS (Exercise for Course "DBMS-Implementation")
// Filename:	err.cpp
// Purpose:	Handling of error and warning messages
// Last Change:	29.05.2013
// Language:	C++
// Author:	Stefan Brass
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2011-2013 by Stefan Brass
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "check.h"
#include "err.h"

#pragma GCC diagnostic ignored "-Wunreachable-code"
#include <iostream>
#include <fstream>

//=============================================================================
// Local Variables of this Module:
//=============================================================================

//-----------------------------------------------------------------------------
// File: The output stream for the alert file.
//-----------------------------------------------------------------------------

static std::ofstream Alert_File;


//=============================================================================
// Exported Functions (Public Static Members):
//=============================================================================


//=============================================================================
// Possibility to query last error messages:
//=============================================================================

//-----------------------------------------------------------------------------
// reset(): Clear buffer of error messages.
//-----------------------------------------------------------------------------

void err_c::reset()
{
	NumErrs = 0;
}

//-----------------------------------------------------------------------------
// ok(): No errors after last reset().
//-----------------------------------------------------------------------------

bool err_c::ok()
{
	return (NumErrs == 0);
}

//-----------------------------------------------------------------------------
// num_errs(): Number of errors after last reset().
//-----------------------------------------------------------------------------

int err_c::num_errs()
{
	return NumErrs;
}

//-----------------------------------------------------------------------------
// msg(): Return n-th/first error message after last reset().
//-----------------------------------------------------------------------------

str_t err_c::msg(int n)
{
	// The parameter must be non-negative:
	CHECK(n >= 0, "err_c::msg: Error index is negative");

	// The parameter must be smaller than the number of messages:
	CHECK(n < NumErrs, "err_c::msg: Error index too large");

	// Return error message:
	return ErrMsg[n];
}

//=============================================================================
// Automatic printing of error messages (Alert-File and/or stderr):
//=============================================================================

//-----------------------------------------------------------------------------
// open_alert: Open the alert file.
//-----------------------------------------------------------------------------

bool err_c::open_alert(str_t filename)
{
	// The file must be closed:
	CHECK(!Alert_File.is_open(),
		"err_c::open_alert: File is already open");

	// Open the file:
	Alert_File.open(filename, std::ios::out|std::ios::app);

	// Return true if file open succeeded:
	return Alert_File.good() && Alert_File.is_open();
}

//-----------------------------------------------------------------------------
// close_alert: Close the alert file.
//-----------------------------------------------------------------------------

bool err_c::close_alert()
{
	// Close alert file:
	Alert_File.close();

	//Check if succeeded:
	return Alert_File.good();
}

//-----------------------------------------------------------------------------
// use_stderr: Define whether error messages are printed to stderr.
//-----------------------------------------------------------------------------

void err_c::use_stderr(bool b)
{
	UseStderr = b;
}

//=============================================================================
// Errors detected in file.cpp:
//=============================================================================


//-----------------------------------------------------------------------------
// open_failed: Error opening a database file.
//-----------------------------------------------------------------------------

void err_c::open_failed(str_t filename, str_t msg)
{
	msg_begin();
	msg_str("Error opening database file '");
	msg_str(filename);
	msg_str("': ");
	msg_str(msg);
	msg_output();
}

//-----------------------------------------------------------------------------
// close_failed: Error closing a database file.
//-----------------------------------------------------------------------------

void err_c::close_failed(str_t filename, str_t msg)
{
	msg_begin();
	msg_str("Error closing database file '");
	msg_str(filename);
	msg_str("': ");
	msg_str(msg);
	msg_output();
}

//-----------------------------------------------------------------------------
// read_failed: Error reading from a database file.
//-----------------------------------------------------------------------------

void err_c::read_failed(str_t filename, str_t msg)
{
	msg_begin();
	msg_str("Error reading from database file '");
	msg_str(filename);
	msg_str("': ");
	msg_str(msg);
	msg_output();
}

//-----------------------------------------------------------------------------
// write_failed: Error writing to a database file.
//-----------------------------------------------------------------------------

void err_c::write_failed(str_t filename, str_t msg)
{
	msg_begin();
	msg_str("Error writing to database file '");
	msg_str(filename);
	msg_str("': ");
	msg_str(msg);
	msg_output();
}

//-----------------------------------------------------------------------------
// seek_failed: Error doing lseek for a database file.
//-----------------------------------------------------------------------------

void err_c::seek_failed(str_t filename, str_t msg)
{
	msg_begin();
	msg_str("Error when setting position in database file '");
	msg_str(filename);
	msg_str("': ");
	msg_str(msg);
	msg_output();
}

//-----------------------------------------------------------------------------
// sync_failed: Error doing fdatasync/fsync for a database file.
//-----------------------------------------------------------------------------

void err_c::sync_failed(str_t filename, str_t msg)
{
	msg_begin();
	msg_str("Error when flushing buffers (sync) for file '");
	msg_str(filename);
	msg_str("': ");
	msg_str(msg);
	msg_output();
}

//-----------------------------------------------------------------------------
// bad_filesize: Real file size is different from file size in header.
//-----------------------------------------------------------------------------

void err_c::bad_filesize(str_t filename, int size, fsize_t blocks)
{
	msg_begin();
	msg_str("Size of file '");
	msg_str(filename);
	msg_str("' (");
	msg_int(size);
	msg_str(" Bytes) does not match size in header (");
	msg_uint(blocks);
	msg_str(" blocks)");
	msg_output();
}

//-----------------------------------------------------------------------------
// bad_block: Data block read from file violates validity check.
//-----------------------------------------------------------------------------

void err_c::bad_block(str_t filename, bno_t block_no, str_t msg)
{
	msg_begin();
	msg_str("Bad block in database file '");
	msg_str(filename);
	msg_str("', block ");
	msg_uint(block_no);
	msg_str(": ");
	msg_str(msg);
	msg_output();
}

//=============================================================================
// Errors detected in buf.cpp:
//=============================================================================


//-----------------------------------------------------------------------------
// new_failed: Error when trying to allocate memory for a block buffer.
//-----------------------------------------------------------------------------

void err_c::new_failed(int num_bufs)
{
	msg_begin();
	msg_str("Error when trying to allocate memory for buffer (");
	msg_int(num_bufs);
	msg_str(" allocated so far)");
	msg_output();
}

//-----------------------------------------------------------------------------
// no_bufs: No unpinned buffers available.
//-----------------------------------------------------------------------------

void err_c::no_bufs(int num_bufs)
{
	msg_begin();
	msg_str("No unpinned buffers available (");
	msg_int(num_bufs);
	msg_str(" allocated)");
	msg_output();
}


//-----------------------------------------------------------------------------
// no_data_block: Block is not of expected type "relation data".
//-----------------------------------------------------------------------------

void err_c::no_data_block(str_t filename, bno_t block_no)
{
	msg_begin();
	msg_str("Block is not of expected type 'Relation Data' (File \"");
	msg_str(filename);
	msg_str("\", block ");
	msg_uint(block_no);
	msg_str(")");
	msg_output();
}


//=============================================================================
// Errors related to segments (fhblk.cpp, seg.cpp):
//=============================================================================

//-----------------------------------------------------------------------------
// insufficient_space: Not enough space in database file.
//-----------------------------------------------------------------------------

void err_c::insufficient_space(str_t filename,
				fsize_t needed_blocks, fsize_t free_blocks)
{
	msg_begin();
	msg_str("Not enough space in \"");
	msg_str(filename);
	msg_str("\" (need ");
	msg_uint(needed_blocks);
	msg_str(" blocks, currently free: ");
	msg_uint(free_blocks);
	msg_str(" blocks)");
	msg_output();
}

//-----------------------------------------------------------------------------
// file_header_full: No more space for segment/extent data.
//-----------------------------------------------------------------------------

void err_c::file_header_full(str_t filename)
{
	msg_begin();
	msg_str("The file header in \"");
	msg_str(filename);
	msg_str("\" is full: ");
	msg_str("Not enough space for segment/extent data");
	msg_output();
}

//-----------------------------------------------------------------------------
// seg_does_not_exist: Trying operation on invalid seg id.
//-----------------------------------------------------------------------------

void err_c::seg_does_not_exist(str_t filename, int seg_id)
{
	msg_begin();
	msg_str("Trying to access segment number ");
	msg_int(seg_id);
	msg_str("in \"");
	msg_str(filename);
	msg_str("\": ");
	msg_str("Segment does not exist");
	msg_output();
}

//-----------------------------------------------------------------------------
// rel_is_full: Insertion of a row has failed, no unused block.
//-----------------------------------------------------------------------------

void err_c::rel_is_full(str_t filename, str_t rname, fsize_t size)
{
	msg_begin();
	msg_str("Relation \"");
	msg_str(rname);
	msg_str("\" in file \"");
	msg_str(filename);
	msg_str("\" is full: Insertion failed");
	msg_str(" (current size: ");
	msg_int(size);
	msg_str(" blocks)");
	msg_output();
}

//-----------------------------------------------------------------------------
// block_outside_segment: Trying to access n-th block, n > size.
//-----------------------------------------------------------------------------

void err_c::block_outside_segment(str_t filename, int seg_id, sbno_t n)
{
	msg_begin();
	msg_str("File \"");
	msg_str(filename);
	msg_str("\", segment ");
	msg_int(seg_id);
	msg_str(": Tried to access block ");
	msg_uint(n);
	msg_str(" outside segment");
	msg_output();
}

//=============================================================================
// Errors related to the data dictionary:
//=============================================================================

//-----------------------------------------------------------------------------
// invalid_seg_id: Invalid segment number in SYSRELS.
//-----------------------------------------------------------------------------

void err_c::invalid_seg_id(str_t filename, int seg_id)
{
	msg_begin();
	msg_str("Invalid segment ID ");
	msg_int(seg_id);
	msg_str(" in SYSRELS (File \"");
	msg_str(filename);
	msg_str("\")");
	msg_output();
}


//-----------------------------------------------------------------------------
// no_dict_block: Error getting data dictionary block.
//-----------------------------------------------------------------------------

void err_c::no_dict_block(str_t filename, bno_t block_no)
{
	msg_begin();
	msg_str("Could not get data dictionary block (File \"");
	msg_str(filename);
	msg_str("\", block ");
	msg_uint(block_no);
	msg_str(")");
	msg_output();
}


//-----------------------------------------------------------------------------
// dict_full: Data dictionary table is full.
//-----------------------------------------------------------------------------

void err_c::dict_full(str_t dict_table, str_t new_table)
{
	msg_begin();
	msg_str("Data dictionary table \"");
	msg_str(dict_table);
	msg_str("\" has not enough space for creating \"");
	msg_str(new_table);
	msg_str("\"");
	msg_output();
}


//-----------------------------------------------------------------------------
// rel_exists_already: Relation could not be created, name clash.
//-----------------------------------------------------------------------------

void err_c::rel_exists_already(str_t rel_name)
{
	msg_begin();
	msg_str("Relation \"");
	msg_str(rel_name);
	msg_str("\" could not be created: ");
	msg_str("there is already a relation with this name");
	msg_output();
}

//-----------------------------------------------------------------------------
// rel_not_found: Relation could not be opened, does not exist.
//-----------------------------------------------------------------------------

void err_c::rel_not_found(str_t rel_name)
{
	msg_begin();
	msg_str("Relation \"");
	msg_str(rel_name);
	msg_str("\" could not be opened: ");
	msg_str("no relation with this name");
	msg_output();
}

//-----------------------------------------------------------------------------
// col_not_found: Relation does not have a requested column.
//-----------------------------------------------------------------------------

void err_c::col_not_found(str_t rel_name, str_t col_name)
{
	msg_begin();
	msg_str("Relation \"");
	msg_str(rel_name);
	msg_str("\" does not have the requested column \"");
	msg_str(col_name);
	msg_str("\"");
	msg_output();
}

//-----------------------------------------------------------------------------
// wrong_coltype: Type in Data Dictionary does not match program.
//-----------------------------------------------------------------------------

void err_c::wrong_coltype(str_t rel_name, str_t col_name)
{
	msg_begin();
	msg_str("Type for relation \"");
	msg_str(rel_name);
	msg_str("\", column \"");
	msg_str(col_name);
	msg_str("\" is differs between database and program");
	msg_output();
}

//=============================================================================
// Functions for Dumping Data:
//=============================================================================

//-----------------------------------------------------------------------------
// Begin dump section in alert file:
//-----------------------------------------------------------------------------

void err_c::dump_begin(str_t headline)
{
	// Check parameter:
	CHECK(headline != 0, "err_c::dump_begin: headline is NULL");

	// Print headline in "----" block:
	dump_output("\n\n");
	dump_output("----------------------------------------");
	dump_output("---------------------------------------\n");
	dump_output(headline);
	dump_output("\n");
	dump_output("----------------------------------------");
	dump_output("---------------------------------------\n");
	dump_output("\n");
}

//-----------------------------------------------------------------------------
// Print a string:
//-----------------------------------------------------------------------------

void err_c::dump_str(str_t str)
{
	if(str == 0)
		dump_output("<NULL>");
	else
		dump_output(str);
}

//-----------------------------------------------------------------------------
// Print a line end:
//-----------------------------------------------------------------------------

void err_c::dump_nl()
{
	dump_output("\n");
}

//-----------------------------------------------------------------------------
// Print an integer in decimal notation:
//-----------------------------------------------------------------------------

void err_c::dump_int(int n)
{
	char buf[STR_INT_LEN];

	// Do the converstion. str_int returns number of characters written:
	int num_chars = str_int(buf, STR_INT_LEN, n);
	CHECK(num_chars < STR_INT_LEN,
		"err_c::dump_int: str_int caused buffer overflow");

	// Print result:
	if(num_chars < 1)
		dump_output("<Error when converting int>");
	else
		dump_output(buf);
}

//-----------------------------------------------------------------------------
// Print an unsigned integer in decimal notation:
//-----------------------------------------------------------------------------

void err_c::dump_uint(unsigned int n)
{
	char buf[STR_INT_LEN];

	// Do the converstion. str_int returns number of characters written:
	int num_chars = str_uint(buf, STR_INT_LEN, n);
	CHECK(num_chars < STR_INT_LEN,
		"err_c::dump_uint: str_uint caused buffer overflow");

	// Print result:
	if(num_chars < 1)
		dump_output("<Error when converting unsigned int>");
	else
		dump_output(buf);
}

//-----------------------------------------------------------------------------
// End dump section in alert file:
//-----------------------------------------------------------------------------

void err_c::dump_end()
{
	// Print message:
	dump_output("\n(end of dump)\n");

	// Flush data:
	dump_flush();
}

//=============================================================================
// Auxillary Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// msg_begin: Begin construction of message in buffer.
//-----------------------------------------------------------------------------

void err_c::msg_begin()
{
	BufPtr = Buf;
	BufFree = ERR_BUFSIZE;
}

//-----------------------------------------------------------------------------
// msg_str: Add/Append string to message buffer.
//-----------------------------------------------------------------------------

void err_c::msg_str(str_t str)
{
	while(*str && --BufFree > 0) {
		// Note that BufFree might become 0 here, although there is
		// still one byte free for the final '\0'.
		CHECK(BufPtr >= Buf, "err_c::msg_str: BufPtr < Buf");
		CHECK(BufPtr < Buf+ERR_BUFSIZE,
			"err_c::msg_str: BufPtr out of Buf");

		*BufPtr++ = *str++;
	}
}

//-----------------------------------------------------------------------------
// msg_int: Add/Append int to message buffer.
//-----------------------------------------------------------------------------

void err_c::msg_int(int i)
{
	// Check buffer pointer:
	CHECK(BufPtr >= Buf, "err_c::msg_int: BufPtr < Buf");
	CHECK(BufPtr < Buf+ERR_BUFSIZE,
			"err_c::msg_int: BufPtr out of Buf");

	// str_int does the work:
	if(BufFree > 1) {
		int n = str_int(BufPtr, BufFree-1, i);
		CHECK(n < BufFree, "err_c::msg_int: bad result of str_int");
		BufFree -= n;
		BufPtr += n;
	}
}

//-----------------------------------------------------------------------------
// msg_uint: Add/Append unsigned int to message buffer.
//-----------------------------------------------------------------------------

void err_c::msg_uint(unsigned int i)
{
	// Check buffer pointer:
	CHECK(BufPtr >= Buf, "err_c::msg_uint: BufPtr < Buf");
	CHECK(BufPtr < Buf+ERR_BUFSIZE,
			"err_c::msg_uint: BufPtr out of Buf");

	// str_uint does the work:
	if(BufFree > 1) {
		int n = str_uint(BufPtr, BufFree-1, i);
		CHECK(n < BufFree, "err_c::msg_uint: bad result of str_int");
		BufFree -= n;
		BufPtr += n;
	}
}

//-----------------------------------------------------------------------------
// msg_output: Output contents of message buffer.
//-----------------------------------------------------------------------------

void err_c::msg_output()
{
	// Check buffer pointer:
	CHECK(BufPtr >= Buf, "err_c::msg_output: BufPtr < Buf");
	CHECK(BufPtr < Buf+(ERR_BUFSIZE-1),
			"err_c::msg_output: BufPtr out of Buf");

	// Terminate string in buffer:
	*BufPtr = '\0';

	// Store Message in error message buffer if there is still space:
	if(NumErrs < ERR_MAXERRS - 1) {
		str_cpy(ErrMsg[NumErrs++], ERR_BUFSIZE, Buf);
	}

	// Print message to alert file if open:
	if(Alert_File.is_open()) {
		Alert_File << Buf << ".\n";
		Alert_File << std::flush;
	}

	// Print message to standard error channel if requested:
	if(UseStderr) {
		std::cerr << "\n" << Buf << ".\n";
		std::cerr << std::flush;
	}
}

//-----------------------------------------------------------------------------
// dump_output: Output a string as part of a data structure dump.
//-----------------------------------------------------------------------------

void err_c::dump_output(str_t str)
{
	// Check buffer pointer:
	CHECK(str != 0, "err_c::dump_output: str is NULL");

	// Print string to alert file if open:
	if(Alert_File.is_open())
		Alert_File << str;

	// Print message to standard error channel if requested:
	if(UseStderr)
		std::cerr << str;
}

//-----------------------------------------------------------------------------
// dump_flush: Make sure that dump data is really written.
//-----------------------------------------------------------------------------

void err_c::dump_flush()
{
	// Flush alert file if output was written there:
	if(Alert_File.is_open())
		Alert_File << std::flush;

	// Flush standard error channel if output was written there:
	if(UseStderr)
		std::cerr << std::flush;
}

//=============================================================================
// Class Attributes:
//=============================================================================


//-----------------------------------------------------------------------------
// UseStderr: Defines whether error messages are printed to stderr.
//-----------------------------------------------------------------------------

bool err_c::UseStderr;

//-----------------------------------------------------------------------------
// Buf: Buffer for current message.
//-----------------------------------------------------------------------------

char err_c::Buf[ERR_BUFSIZE];

//-----------------------------------------------------------------------------
// BufPtr: Next free position in buffer.
//-----------------------------------------------------------------------------

char *err_c::BufPtr;

//-----------------------------------------------------------------------------
// BufFree: Number of free characters in buffer.
//-----------------------------------------------------------------------------

int err_c::BufFree;


//-----------------------------------------------------------------------------
// NumErrs: Number of error messages after last reset().
//-----------------------------------------------------------------------------

int err_c::NumErrs = 0;

//-----------------------------------------------------------------------------
// ErrMsg: Array of last error messages.
//-----------------------------------------------------------------------------

char err_c::ErrMsg[ERR_MAXERRS][ERR_BUFSIZE];

