// ============================================================================
// Project:	A Tiny DBMS (Exercise for Course "DBMS-Implementation")
// Filename:	err.h
// Purpose:	Handling of error and warning messages
// Last Change:	29.05.2013
// Language:	C++
// Author:	Stefan Brass
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2013 by Stefan Brass
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	ERR_INCLUDED
#define ERR_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef BNO_INCLUDED
#include "bno.h"
#endif

#ifndef SBNO_INCLUDED
#include "sbno.h"
#endif

#ifndef FSIZE_INCLUDED
#include "fsize.h"
#endif

class ofstream;

//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// ERR_BUFSIZE: Size of buffer for a single error message.
//-----------------------------------------------------------------------------

static const int ERR_BUFSIZE = 128;

//-----------------------------------------------------------------------------
// ERR_MAXERRS: Maximal number of error messages kept in memory.
//-----------------------------------------------------------------------------

static const int ERR_MAXERRS = 10;

//=============================================================================
// Module for Error Message Handling (Class with only static members):
//=============================================================================

class err_c {
	public:


//-----------------------------------------------------------------------------
// Static Methods:
//-----------------------------------------------------------------------------

	//---------------------------------------------------------------------
	// Possibility to query last error messages:
	//---------------------------------------------------------------------

	// reset(): Clear buffer of error messages.
	static void reset();

	// ok(): No errors after last reset().
	static bool ok();

	// num_errs(): Number of errors after last reset().
	static int num_errs();

	// msg(): Return n-th/first error message after last reset().
	static str_t msg(int n = 0);

	//---------------------------------------------------------------------
	// Automatic printing of error messages (Alert-File and/or stderr):
	//---------------------------------------------------------------------

	// open: Open the alert file.
	static bool open_alert(str_t filename);

	// close: Close the alert file.
	static bool close_alert();

	// use_stderr: Define whether error messages are printed to stderr:
	static void use_stderr(bool b);

	//---------------------------------------------------------------------
	// Errors detected in file.cpp:
	//---------------------------------------------------------------------

	// open_failed: Error opening a database file.
	static void open_failed(str_t filename, str_t msg);

	// close_failed: Error closing a database file.
	static void close_failed(str_t filename, str_t msg);

	// read_failed: Error reading from a database file.
	static void read_failed(str_t filename, str_t msg);

	// write_failed: Error writing to a database file.
	static void write_failed(str_t filename, str_t msg);

	// seek_failed: Error doing lseek for a database file.
	static void seek_failed(str_t filename, str_t msg);

	// sync_failed: Error doing fdatasync/fsync for a database file.
	static void sync_failed(str_t filename, str_t msg);

	// bad_filesize: Real file size is different from file size in header.
	static void bad_filesize(str_t filename, int size, fsize_t blocks);

	// bad_block: Data block read from file violates validity check.
	static void bad_block(str_t filename, bno_t block_no, str_t msg);

	//---------------------------------------------------------------------
	// Errors detected in buf.cpp:
	//---------------------------------------------------------------------

	// new_failed: Error when trying to allocate memory for a block buffer.
	static void new_failed(int num_bufs);

	// no_bufs: No unpinned buffers available.
	static void no_bufs(int num_bufs);

	// no_data_block: Block is not of expected type "relation data".
	static void no_data_block(str_t filename, bno_t block_no);

	//---------------------------------------------------------------------
	// Errors related to segments (fhblk.cpp, seg.cpp):
	//---------------------------------------------------------------------

	// insufficient_space: Not enough space in database file.
	static void insufficient_space(str_t filename,
				fsize_t needed_blocks, fsize_t free_blocks);

	// file_header_full: No more space for segment/extent data.
	static void file_header_full(str_t filename);

	// seg_does_not_exist: Trying operation on invalid seg id.
	static void seg_does_not_exist(str_t filename, int seg_id);

	// rel_is_full: Insertion of a row has failed, no unused block.
	static void rel_is_full(str_t filename, str_t rname, fsize_t size);

	// block_outside_segment: Trying to access n-th block, n > size.
	static void block_outside_segment(str_t filename, int seg_id,
			sbno_t n);

	//---------------------------------------------------------------------
	// Errors related to the data dictionary:
	//---------------------------------------------------------------------

	// invalid_seg_id: Invalid segment number in SYSRELS.
	static void invalid_seg_id(str_t filename, int seg_id);

	// no_dict_block: Error getting data dictionary block.
	static void no_dict_block(str_t filename, bno_t block_no);

	// dict_full: Data dictionary table is full.
	static void dict_full(str_t dict_table, str_t new_table);

	// rel_exists_already: Relation could not be created, name clash.
	static void rel_exists_already(str_t rel_name);

	// rel_not_found: Relation could not be opened, does not exist.
	static void rel_not_found(str_t rel_name);

	// col_not_found: Relation does not have a requested column.
	static void col_not_found(str_t rel_name, str_t col_name);

	// wrong_coltype: Type in Data Dictionary does not match program.
	static void wrong_coltype(str_t rel_name, str_t col_name);

	//---------------------------------------------------------------------
	// Functions for Dumping Data:
	//---------------------------------------------------------------------

	// Begin dump section in alert file:
	static void dump_begin(str_t headline);

	// Print a string:
	static void dump_str(str_t str);

	// Print a line end:
	static void dump_nl();

	// Print an integer in decimal notation:
	static void dump_int(int n);

	// Print an unsigned integer in decimal notation:
	static void dump_uint(unsigned int n);

	// End dump section in alert file:
	static void dump_end();

//-----------------------------------------------------------------------------
// Auxillary Static Methods:
//-----------------------------------------------------------------------------

	private:

	// msg_begin: Begin construction of message in buffer.
	static void msg_begin();

	// msg_str: Add/Append string to message buffer.
	static void msg_str(str_t str);

	// msg_int: Add/Append int to message buffer.
	static void msg_int(int i);

	// msg_uint: Add/Append unsigned int to message buffer.
	static void msg_uint(unsigned int i);

	// msg_output: Output contents of message buffer.
	static void msg_output();

	// dump_output: Output a string as part of a data structure dump.
	static void dump_output(str_t str);

	// dump_flush: Make sure that dump data is really written.
	static void dump_flush();

//-----------------------------------------------------------------------------
// Class Variables:
//-----------------------------------------------------------------------------

	private:

	// UseStderr: Defines whether error messages are printed to stderr.
	static bool UseStderr;

	// Buf: Buffer for current message.
	static char Buf[ERR_BUFSIZE];

	// BufPtr: Next free position in buffer.
	static char *BufPtr;

	// BufFree: Number of free characters in buffer.
	static int BufFree;

	// NumErrs: Number of error messages after last reset().
	static int NumErrs;

	// ErrMsg: Array of last error messages.
	static char ErrMsg[ERR_MAXERRS][ERR_BUFSIZE];

//-----------------------------------------------------------------------------
// End of Class:
//-----------------------------------------------------------------------------

};


//=============================================================================
// End of Include File:
//=============================================================================

#endif

