// ============================================================================
// Project:	A Tiny DBMS (Exercise for Course "DBMS-Implementation")
// Filename:	fhblk.h
// Purpose:	File Header Block
// Last Change:	17.02.2012
// Language:	C++
// Author:	Stefan Brass
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2011-2012 by Stefan Brass
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	FHBLK_INCLUDED
#define FHBLK_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef BNO_INCLUDED
#include "bno.h"
#endif

#ifndef FSIZE_INCLUDED
#include "fsize.h"
#endif

#ifndef BOFF_INCLUDED
#include "boff.h"
#endif

#ifndef BLOCK_INCLUDED
#include "block.h"
#endif


//=============================================================================
// Structure of File Header Block:
//=============================================================================

// No new attributes can be introduced in subclasses of block_c,
// because all blocks must have the same size (VER_BLOCKSIZE).
// However, the Data part of the superclass block_c can be interpreted.
//
// We assume that an int is two shorts long.
// E.g. Data.IntData[0] is the same as Data.ShortData[0] and Data.ShortData[1].
//
// Data.IntData[0]: Number of Blocks in the File
#define FHBLK_NUM_BLOCKS	(Data.UIntData[0])

// Data.IntData[1]: Total number of unused (free) blocks in file.
#define FHBLK_NUM_FREE		(Data.UIntData[1])

// Version number of the file (increased for each checkpoint):
#define FHBLK_FVERSION		(Data.IntData[2])


//=============================================================================
// Class for File Header Block:
//=============================================================================

class fhblk_c : public block_c {

	public:

//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	// Dummy constructor, the real initialization happens in init():
	fhblk_c();

	// The destructor marks the block as unusable:
	~fhblk_c();

//-----------------------------------------------------------------------------
// (Object) Methods:
//-----------------------------------------------------------------------------

	// The validity/integrity check is needed also in non-debug mode.
	// Blocks read from the disk must be checked.
	// A null pointer is returned if everything is ok, else an error msg.
	str_t	invalid() const;

	// Initialize header block (store size, reset possibly used block):
	void init(fsize_t size);

	// Number of blocks in file (file size):
	inline fsize_t num_blocks() const
		{ CHECK_VALID("fhblk_c::num_blocks");
		  return(FHBLK_NUM_BLOCKS); }

	// Set number of blocks in file (file size):
	inline void set_num_blocks(fsize_t nblks)
		{ CHECK_VALID("fhblk_c::set_num_blocks");
		  CHECK(nblks >= 2, "fhblk_c:set_num_blocks: Must be >= 2");
		  FHBLK_NUM_BLOCKS = nblks; }

	// Total number of free blocks in file:
	inline fsize_t num_free() const
		{ CHECK_VALID("fhblk_c::num_free");
		  return(FHBLK_NUM_FREE); }

	// Write contents of file header block in a readable way to alert file.
	void dump(str_t headline);

//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	fhblk_c(const fhblk_c& fhblk);			// Not implemented
	fhblk_c& operator=(const fhblk_c& fhblk);	// Not implemented


//-----------------------------------------------------------------------------
// End of Class:
//-----------------------------------------------------------------------------

};

//-----------------------------------------------------------------------------
// Define pointer type:
//-----------------------------------------------------------------------------

typedef fhblk_c *fhblk_t;

//=============================================================================
// End of Include File:
//=============================================================================

#endif

