// ============================================================================
// Project:	A Tiny DBMS (Exercise for Course "DBMS-Implementation")
// Filename:	file.cpp
// Purpose:	Interface to Data Files (for reading and writing blocks)
// Last Change:	21.02.2012
// Language:	C++
// Author:	Stefan Brass
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2011-2012 by Stefan Brass
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "check.h"
#include "bno.h"
#include "fsize.h"
#include "err.h"
#include "file.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

//=============================================================================
// Constructor, Destructor:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor (with Filename):
//-----------------------------------------------------------------------------

file_c::file_c(str_t par_filename)
	: FileHeader()
{
	// Check parameters:
	CHECK(par_filename != 0, "file_c::constructor: filename is NULL");

	// Initialize attributes:
	CHECK_CODE(Magic = FILE_MAGIC);
	Filename = par_filename;
	Hashcode = FileNextHashcode++;
	FD = -1;
	Size = 0;
	Modified = false;
	FileVersion = 0;
}

//-----------------------------------------------------------------------------
// Destructor (file must be closed when destructor is called):
//-----------------------------------------------------------------------------

file_c::~file_c()
{
	// The file must be closed (else file header block maybe not saved):
	if(FD != -1) {
		close();
	}

	// Make this object unusable:
#ifdef	VER_DEBUG
	Magic = 0;
#endif

}


//=============================================================================
// Exported (Object) Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// create: Create this file with a given number of blocks.
//-----------------------------------------------------------------------------

bool	file_c::create(fsize_t num_blocks)
{
	block_c empty_block;

	// Check parameter:
	CHECK(num_blocks >= 4, "file_c::create: num_blocks must be >= 4");
		// Header block, SYSRELS and SYSCOLS use three blocks.
		// There should be at least one block for data.

	// Check object:
	CHECK_VALID("file_c::create");
	
	// Check object state:
	CHECK(FD == -1, "file_c::create: file must be closed");

	// Open file (delete contents if exists):
	FD = ::open(Filename, O_WRONLY|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR);
	if(FD < 0) {
		err_c::open_failed(Filename, strerror(errno));
		return false;
	}

	// Format header block:
	FileHeader.init(num_blocks);

	// Initialize data dictionary:
	// ... later
	
	// Now write the header block:
	int n = ::write(FD, static_cast<void*>(&FileHeader), VER_BLOCKSIZE);
	if(n != VER_BLOCKSIZE) {
		err_c::write_failed(Filename, strerror(errno));
		if(::close(FD) != 0)
			err_c::close_failed(Filename, strerror(errno));
		FD = -1;
		return false;
	}

	// Now write num_blocks-1 copies of the empty block:
	for(bno_t block_no = 1; block_no < num_blocks; block_no++) {
		empty_block.set_block_no(block_no);
		n = ::write(FD, static_cast<void*>(&empty_block),
				VER_BLOCKSIZE);
		if(n != VER_BLOCKSIZE) {
			err_c::write_failed(Filename, strerror(errno));
			if(::close(FD) != 0)
				err_c::close_failed(Filename, strerror(errno));
			FD = -1;
			return false;
		}
	}

	// Close file (we don't call close() because file was not open()ed:
	if(::close(FD) != 0) {
		err_c::close_failed(Filename, strerror(errno));
		FD = -1;
		return false;
	}
	FD = -1;

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// open: Open this file for reading and writing.
//-----------------------------------------------------------------------------

bool	file_c::open()
{
	// Check object:
	CHECK_VALID("file_c::open");
	
	// Check object state:
	CHECK(FD == -1, "file_c::open: file must be closed");

	// Open file:
	FD = ::open(Filename, O_RDWR);
	if(FD < 0) {
		err_c::open_failed(Filename, strerror(errno));
		return false;
	}

	// Read File Header Block:
	if(!read(0, &FileHeader))
		return false;

	// Position to end of file:
	off_t n = lseek(FD, 0, SEEK_END);
	if(n < 0) {
		err_c::seek_failed(Filename, strerror(errno));
		return false;
	}

	// Check current file size:
	fsize_t num_blocks = FileHeader.num_blocks();
	if(n != static_cast<off_t>(num_blocks) * VER_BLOCKSIZE) {
		err_c::bad_filesize(Filename, n, num_blocks);
	}

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// close: Close this file.
//-----------------------------------------------------------------------------

bool	file_c::close()
{
	bool status = true;

	// Check object:
	CHECK_VALID("file_c::close");
	
	// Check object state:
	CHECK(FD >= 0, "file_c::close: file must be open");

	// If the file is modified, the header block must be written:
	if(Modified) {
		if(!write_header())
			status = false;
	}

	// Close file:
	if(::close(FD) != 0) {
		err_c::close_failed(Filename, strerror(errno));
		status = false;
	}

	// Mark file as closed:
	FD = -1;

	// Return true if everything was successful:
	return status;
}


//-----------------------------------------------------------------------------
// read: Read a block.
//-----------------------------------------------------------------------------

bool	file_c::read(bno_t block_no, block_t block)
{
	// Check parameter:
	CHECK(block != 0, "file_c::read: block is NULL");
	CHECK_PAR(block, "file_c::read");

	// Check object:
	CHECK_VALID("file_c::read");
	
	// Check object state:
	CHECK(FD >= 0, "file_c::read: file must be open");

	// Set current position in file:
	fsize_t offset = block_no * VER_BLOCKSIZE;
	off_t n = lseek(FD, offset, SEEK_SET);
	if(n < 0 || static_cast<fsize_t>(n) != offset) {
		err_c::seek_failed(Filename, strerror(errno));
		return false;
	}

	// Read one block:
	n = ::read(FD, static_cast<void*>(block), VER_BLOCKSIZE);
	if(n != VER_BLOCKSIZE) {
		err_c::read_failed(Filename, strerror(errno));
		return false;
	}

	// When a block is read, it is always checked for validity:
	str_t msg = block->invalid();
	if(msg) {
		err_c::bad_block(Filename, block_no, msg);
		return false;
	}

	// We also check the block number in particular:
	if(block->block_no() != block_no) {
		err_c::bad_block(Filename, block_no,
				"wrong block number");
		return false;
	}

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// write: Write a block.
//-----------------------------------------------------------------------------

bool	file_c::write(bno_t block_no, block_t block)
{
	// Check parameter:
	CHECK(block != 0, "file_c::read: block is NULL");
	CHECK_PAR(block, "file_c::read");

	// Check object:
	CHECK_VALID("file_c::write");
	
	// Check object state:
	CHECK(FD >= 0, "file_c::write: file must be open");

	// The block number of the block must be set correctly:
	CHECK(block_no == block->block_no(),
		"file_c::write: Trying to write block at wrong position");

	// Set current position in file:
	off_t offset = block_no * static_cast<off_t>(VER_BLOCKSIZE);
	off_t n = lseek(FD, offset, SEEK_SET);
	if(n != offset) {
		err_c::seek_failed(Filename, strerror(errno));
		return false;
	}

	// Write one block:
	n = ::write(FD, static_cast<void*>(block), VER_BLOCKSIZE);
	if(n != VER_BLOCKSIZE) {
		err_c::write_failed(Filename, strerror(errno));
		return false;
	}

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// sync: Ensure that written blocks are really written to disk.
//-----------------------------------------------------------------------------

bool	file_c::sync()
{
	// Check object:
	CHECK_VALID("file_c::sync");
	
	// Check object state:
	CHECK(FD >= 0, "file_c::sync: file must be open");

	// Call OS sync function:
	int n = -1;
#if _POSIX_C_SOURCE >= 199309L || _XOPEN_SOURCE >= 500
	n = fdatasync(FD);
#elif  _BSD_SOURCE || _XOPEN_SOURCE || _POSIX_C_SOURCE >= 200112L
	n = fsync(FD);
#else
	CHECK_IMPOSSIBLE("file_c::sync: No fdatasync and no fsync");
#endif

	// Check return value:
	if(n != 0) {
		err_c::sync_failed(Filename, strerror(errno));
		return false;
	}

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// extend: Make this file longer by a given number of blocks.
//-----------------------------------------------------------------------------

bool	file_c::extend(fsize_t num_blocks)
{
	block_c empty_block;

	// Check parameter:
	CHECK(num_blocks >= 1,
		"file_c::extend: num_blocks must be at least 1");

	// Check object:
	CHECK_VALID("file_c::extend");
	
	// Check object state:
	CHECK(FD >= 0, "file_c::extend: file must be open");

	// Determine start block number (current file size):
	int block_no = FileHeader.num_blocks();

	// Position to end of file:
	off_t n = lseek(FD, 0, SEEK_END);
	if(n < 0) {
		err_c::seek_failed(Filename, strerror(errno));
		return false;
	}

	// Check current file size:
	if(n != block_no * static_cast<off_t>(VER_BLOCKSIZE)) {
		err_c::bad_filesize(Filename, n, block_no);
	}

	// Write num_blocks blocks:
	while(num_blocks-- > 0) {
		empty_block.set_block_no(block_no);
		n = ::write(FD, static_cast<void*>(&empty_block),
				VER_BLOCKSIZE);
		if(n != VER_BLOCKSIZE) {
			err_c::write_failed(Filename, strerror(errno));
			return false;
		}
		block_no++;
	}

	// Set the new file size:
	FileHeader.set_num_blocks(block_no);

	// Mark file as modified (so that header block is written in close()):
	set_modified();

	// Ok:
	return true;
}

//-----------------------------------------------------------------------------
// write_header: Write file header block.
//-----------------------------------------------------------------------------

bool file_c::write_header()
{
	// Check consistency of file header block:
	CHECK_PAR(&FileHeader, "file_c::write_header");

	// Write file header block:
	return write(0, &FileHeader);
}

//-----------------------------------------------------------------------------
// Integrity check (only for debugging purposes):
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG

str_t	file_c::invalid() const
{
	// Check object pointer:
	if(this == 0)
		return "NIL file_c object";

	// Check data members:
	if(Magic != FILE_MAGIC)
		return "invalid file_c object (bad magic number)";
	if(Filename == 0)
		return "file_c::Filename is NULL";

	// No error found:
	return 0;
}

#endif


//=============================================================================
// Class Attributes (Private Static Members):
//=============================================================================

//-----------------------------------------------------------------------------
// FileNextHashcode: Next Hashcode for a File Object (sequential number).
//-----------------------------------------------------------------------------

int file_c::FileNextHashcode = 0;

