// ============================================================================
// Project:	A Tiny DBMS (Exercise for Course "DBMS-Implementation")
// Filename:	file.h
// Purpose:	Interface to Data Files (for reading and writing blocks)
// Last Change:	13.02.2012
// Language:	C++
// Author:	Stefan Brass
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2012 by Stefan Brass
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	FILE_INCLUDED
#define FILE_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef BNO_INCLUDED
#include "bno.h"
#endif

#ifndef FSIZE_INCLUDED
#include "fsize.h"
#endif

#ifndef FHBLK_INCLUDED
#include "fhblk.h"
#endif


//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// FILE_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long FILE_MAGIC = 0x46494C1AL;	// 'FIL\n'

//=============================================================================
// Class for Data Files:
//=============================================================================

class file_c {
	public:


//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	// Constructor with Filename:
	file_c(str_t filename);

	// Destructor (file must be closed when destructor is called):
	~file_c();


//-----------------------------------------------------------------------------
// (Object) Methods:
//-----------------------------------------------------------------------------

	// filename: Get the filename of this file:
	inline str_t filename() const
		{ CHECK_VALID("file_c::filename");
		  return(Filename); }

	// hashcode: Get the hashcode of this file:
	inline int hashcode() const
		{ CHECK_VALID("file_c::hashcode");
		  return(Hashcode); }

	// create: Create this file with a given number of blocks.
	bool create(fsize_t num_blocks);

	// open: Open this file for reading and writing.
	bool open();

	// close: Close this file.
	bool close();

	// read: Read a block.
	bool read(bno_t block_no, block_t block);

	// write: Write a block.
	bool write(bno_t block_no, block_t block);

	// sync: Ensure that written blocks are really written to disk.
	bool sync();

	// extend: Make this file longer by a given number of blocks.
	bool extend(fsize_t num_blocks);

	// Return the file header block:
	inline fhblk_t header()
		{ CHECK_VALID("file_c::header");
		  return &FileHeader;
		}

	// Write file header block:
	bool write_header();

	// Mark the file as modified (increment version number):
	inline void set_modified()
		{ CHECK_VALID("file_c::set_modified");
		  if(!Modified) {
			Modified = true;
			FileVersion++;
		  }
		}

//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	// Integrity check:
	public:
	str_t	invalid() const;

	// Magic number (identifies objects of this class for debugging).
	private:
	long	Magic;		// Must be "FILE_MAGIC".
#endif

//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	file_c(const file_c& file);		// Not implemented
	file_c& operator=(const file_c& file);	// Not implemented

//-----------------------------------------------------------------------------
// Class Attributes (Private Static Members):
//-----------------------------------------------------------------------------

	private:

	// Next Hashcode for a File (incremented whenever file object created).
	static int FileNextHashcode;

//-----------------------------------------------------------------------------
// Attributes (Private Members):
//-----------------------------------------------------------------------------

	private:

	// Filename: The name of this file.
	str_t Filename;

	// Hashcode: Hashcode (sequentail number) of this file object.
	int Hashcode;

	// FD: The file descriptor (-1 if not open).
	int FD;

	// Size: Number of blocks of this file (0 if not known).
	fsize_t Size;

	// The file header block:
	fhblk_c FileHeader;

	// Was the file modified?
	bool Modified;

	// File version number:
	int FileVersion;

//-----------------------------------------------------------------------------
// End of Class:
//-----------------------------------------------------------------------------

};

//-----------------------------------------------------------------------------
// Define pointer type:
//-----------------------------------------------------------------------------

typedef file_c *file_t;

//=============================================================================
// End of Include File:
//=============================================================================

#endif

