// ============================================================================
// Project:	A Tiny DBMS (Exercise for Course "DBMS-Implementation")
// Filename:	main.cpp
// Purpose:	Main program
// Last Change:	21.02.2012
// Language:	C++
// Author:	Stefan Brass
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2011-2012 by Stefan Brass
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "boff.h"
#include "block.h"
#include "err.h"
#include "perf.h"
#include "file.h"

// #pragma GCC diagnostic ignored "-Wunreachable-code"
#include <iostream>
#include <fstream>
using namespace std;


//=============================================================================
// Choose Tests to be Done:
//=============================================================================

#define MAIN_TEST_BLOCK
#define MAIN_TEST_FILE
#define MAIN_TEST_FILE_PERF

//=============================================================================
// Local Functions (Tests):
//=============================================================================

//-----------------------------------------------------------------------------
// main_test_block: Simple Tests for Blocks.
//-----------------------------------------------------------------------------

#ifdef MAIN_TEST_BLOCK

void main_test_block()
{
	// Write Headline:
	cout << "\n";
	cout << "Simple Test of Block Objects:\n";
	cout << "============================\n";

	// Create a block object:
	block_c b;

	// Check size of block:
	cout << "Size of block object: " << sizeof(b) << "\n";

	// Check whether the size of block objects is indeed VER_BLOCKSIZE.
	// In order to avoid the "unreachable code" message of the compiler
	// we must do the comparison such that the result is not known
	// at compile time, i.e. we make it unclear that n == sizeof(b):
	int n = sizeof(b);
	for(int i = 1; i <= 1; i++)
		n *= i;
	if(n != VER_BLOCKSIZE)
		cout << "This is not the intended size!\n";

	// Check whether block is correct:
	str_t err = b.invalid();
	if(err)
		cout << "Block is damaged: " << err << ".\n";
	else
		cout << "Block is ok!\n";

	// Set block number:
	b.set_block_no(1);

	// Exchange partial write indicators:
	b.prepare_write();

	// Query the block number:
	cout << "Block number: " << b.block_no() << "\n";
	if(b.block_no() != 1)
		cout << "This is not the correct block number!\n";

	// Query the block type:
	cout << "Block type: " << btype_name(b.btype()) << "\n";

	// Check whether block is correct:
	err = b.invalid();
	if(err)
		cout << "Block is now damaged: " << err << ".\n";
	else
		cout << "Block is still ok!\n";
}
#endif


//-----------------------------------------------------------------------------
// Test file interface
//-----------------------------------------------------------------------------

#ifdef MAIN_TEST_FILE

static void main_test_file()
{
	// Write Headline:
	cout << "\n";
	cout << "First Test of Data Files (Create, Read/Write Block):\n";
	cout << "===================================================\n";

	// Create block and file object:
	block_c block;
	file_c file1("testfile1");

	// Create and open file with 100 blocks:
	if(!file1.create(100)) {
		cout << "File create failed!\n";
		return;
	}
	if(!file1.open()) {
		cout << "File open failed!\n";
		return;
	}

	// Read block with number 1 (first block after header), check contents:
	if(!file1.read(1, &block)) {
		cout << "File read failed!\n";
		return;
	}
	for(boff_t i = 0; i < BLOCK_DATASIZE; i++) {
		if(block.databyte(i) != 0) {
			cout << "Kein Nullbyte in Block 1 an Position " << i
				<< "!\n";
			return;
		}
	}

	// Fill data portion of block, write it in block position 3:
	for(boff_t i = 0; i < BLOCK_DATASIZE; i++) {
		block.set_databyte(i, "ABCDEFG\n"[i%8]);
	}
	block.set_block_no(3);
	if(!file1.write(3, &block)) {
		cout << "File write failed!\n";
		return;
	}
	if(!file1.sync()) {
		cout << "File sync failed!\n";
		return;
	}

	// Read block again, check data part:
	if(!file1.read(3, &block)) {
		cout << "File read failed!\n";
		return;
	}
	for(boff_t i = 0; i < BLOCK_DATASIZE; i++) {
		if(block.databyte(i) != "ABCDEFG\n"[i%8]) {
			cout << "Block 3, Zeichen an Position " << i
				<< " ist falsch!\n";
			return;
		}
	}

	// Done:
	if(!file1.close()) {
		cout << "File close failed!\n";
		return;
	}

	// Ok:
	cout << "test_file: ok\n";
}

#endif

//-----------------------------------------------------------------------------
// Time file operations:
//-----------------------------------------------------------------------------

#ifdef MAIN_TEST_FILE_PERF

static void main_test_file_perf()
{
	// Write Headline:
	cout << "\n";
	cout << "Measure Time for Creating 80MB File:\n";
	cout << "===================================\n";

	// Create block, file and timer object:
	block_c block;
	file_c file("testfile2");
	perf_c timer;

	// Measure time for creating 80MB file:
	timer.start();
	if(!file.create(10000)) {
		cout << "File create failed!\n";
		return;
	}
	timer.stop();
	cout << "Time for creating file with 10000 blocks: "
		<< timer.realtime_msec() << " ms "
		<< "(" << timer.cputime_msec() << "ms CPU)\n";

	// Fill a block with data:
	for(boff_t i = 0; i < BLOCK_DATASIZE; i++) {
		block.set_databyte(i, "ABCDEFG\n"[i%8]);
	}

	// Measure time for writing 80MB file:
	file.open();
	timer.start();
	for(int block_no = 1; block_no <= 10000; block_no++) {
		block.set_block_no(block_no);
		if(!file.write(block_no, &block)) {
			cout << "File write failed!\n";
			return;
		}
	}
	if(!file.sync()) {
		cout << "File sync failed!\n";
		return;
	}
	timer.stop();
	if(!file.close()) {
		cout << "File close failed!\n";
		return;
	}
	cout << "Time for writing file with 10000 blocks: "
		<< timer.realtime_msec() << " ms "
		<< "(" << timer.cputime_msec() << "ms CPU)\n";
}

#endif


//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// main: Main program.
//-----------------------------------------------------------------------------

int main()
{
	// Open alert file:
	if(!err_c::open_alert(VER_ALERT_FILE)) {
		cerr << "Alert file could not be opened!\n";
		return 1;
	}
	err_c::use_stderr(true);

#ifdef MAIN_TEST_BLOCK
	main_test_block();
#endif

#ifdef MAIN_TEST_FILE
	// Test file_c:
	main_test_file();
#endif

	// Measure times for creating/writing file:
#ifdef MAIN_TEST_FILE_PERF
	main_test_file_perf();
#endif

	// Close alert file:
	if(!err_c::close_alert()) {
		cerr << "Error closing alert file!\n";
		return 1;
	}

	// Terminate with success:
	return 0;
}

