// ============================================================================
// Project:	A Tiny DBMS (Exercise for Course "DBMS-Implementation")
// Filename:	perf.cpp
// Purpose:	Interface to Data Files (for reading and writing blocks)
// Last Change:	01.11.2011
// Language:	C++
// Author:	Stefan Brass
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2011 by Stefan Brass
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "check.h"
#include "perf.h"

//=============================================================================
// Operating System dependent Include Files:
//=============================================================================


#ifdef VER_UNIX
#include <sys/time.h>
#endif

#ifdef VER_WINDOWS
#include <windows.h>
#endif

#include <time.h>


//=============================================================================
// Global Variables of this Module:
//=============================================================================

//-----------------------------------------------------------------------------
// Perf_NumUsed: Number of initialized places in Perf_InUse
//-----------------------------------------------------------------------------

static int Perf_NumUsed = 0;

//-----------------------------------------------------------------------------
// Perf_InUse: Indexes in data arrays that are currently in use.
//-----------------------------------------------------------------------------

static bool Perf_InUse[VER_MAX_TIMERS];

//-----------------------------------------------------------------------------
// Perf_StartRTime, Perf_StopRTime: Data for real time measurement.
//-----------------------------------------------------------------------------

#ifdef VER_UNIX
static timeval Perf_StartRTime[VER_MAX_TIMERS];
static timeval Perf_StopRTime[VER_MAX_TIMERS];
#endif

#ifdef VER_WINDOWS
static union {
		long long ns100;
		FILETIME ft;
	} Perf_StartRTime[VER_MAX_TIMERS], Perf_StopRTime[VER_MAX_TIMERS];
#endif

//-----------------------------------------------------------------------------
// Perf_StartCTime, Perf_StopCTime: Data for CPU time measurement.
//-----------------------------------------------------------------------------

static clock_t Perf_StartCTime[VER_MAX_TIMERS];
static clock_t Perf_StopCTime[VER_MAX_TIMERS];

//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

perf_c::perf_c()
{
	// Find free index (or set ID to -1 if none):
	if(Perf_NumUsed < VER_MAX_TIMERS) {
		ID = Perf_NumUsed;
		Perf_InUse[ID] = true;
		Perf_NumUsed++;
	}
	else {
		ID = -1;
		for(int i = 0; i < VER_MAX_TIMERS; i++) {
			if(!Perf_InUse[i]) {
				ID = i;
				Perf_InUse[ID] = true;
				Perf_NumUsed++;
			}
		}
	}

	// Set Magic number:
	CHECK_CODE(Magic = PERF_MAGIC);

	// Initialize timer data:
	if(ID >= 0) {
		start();
		stop();
	}
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

perf_c::~perf_c()
{
	// Check this object:
	CHECK_VALID("perf_c::destructor");

	// Free entry in arrays:
	if(ID != -1)
		Perf_InUse[ID] = false;

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}

//-----------------------------------------------------------------------------
// start: Start time measurement.
//-----------------------------------------------------------------------------

void perf_c::start()
{
	// Check this object:
	CHECK_VALID("perf_c::start");

	// Remember current time if we got a place in the arrays:
	if(ID != -1) {

		// Current Real Time, UNIX version:
#ifdef VER_UNIX
		gettimeofday(Perf_StartRTime+ID, 0);
#endif

		// Current Real Time, Windows version:
#ifdef VER_WINDOWS
		GetSystemTimeAsFileTime(&(Perf_StartRTime[ID].ft));
#endif

		// Current CPU time:
		Perf_StartCTime[ID] = clock();
	}
}

//-----------------------------------------------------------------------------
// stop: Stop time measurement.
//-----------------------------------------------------------------------------

void perf_c::stop()
{
	// Check this object:
	CHECK_VALID("perf_c::stop");

	// Remember current time if we got a place in the arrays:
	if(ID != -1) {
		// Current Real Time, UNIX version:
#ifdef VER_UNIX
		gettimeofday(Perf_StopRTime+ID, 0);
#endif

		// Current Real Time, Windows version:
#ifdef VER_WINDOWS
		GetSystemTimeAsFileTime(&(Perf_StopRTime[ID].ft));
#endif

		// Current CPU time:
		Perf_StopCTime[ID] = clock();
	}
}

//-----------------------------------------------------------------------------
// realtime_msec: Time between last start() and stop() in milliseconds.
//-----------------------------------------------------------------------------

long int perf_c::realtime_msec()
{
	// Check this object:
	CHECK_VALID("perf_c::realtime_msec");

	// If it was not possible to store timing data, return 0:
	if(ID == -1)
		return 0;

	// Compute milliseconds, UNIX Version:
#ifdef VER_UNIX
	long sec = Perf_StopRTime[ID].tv_sec - Perf_StartRTime[ID].tv_sec;
	long usec = Perf_StopRTime[ID].tv_usec - Perf_StartRTime[ID].tv_usec;
// XXX cout << "sec=" << sec << ", usec= " << usec << "\n";
	if(usec < 0) {
		usec += 1000000;
		sec -= 1;
	}
	return sec * 1000 + (usec + 500) / 1000;
#endif

	// Compute milliseconds, Windows Version:
#ifdef VER_WINDOWS
	return static_cast<long>
			((Perf_StopRTime[ID].ns100 - Perf_StartRTime[ID].ns100)
					/ 10000);
#endif

}

//-----------------------------------------------------------------------------
// cputime_msec: CPU time between last start() and stop() in millisecs.
//-----------------------------------------------------------------------------

long int perf_c::cputime_msec()
{
	// Check this object:
	CHECK_VALID("perf_c::cputime_msec");

	// If it was not possible to store timing data, return 0:
	if(ID == -1)
		return 0;

	// Compute CPU time in milliseconds:
	long ticks = Perf_StopCTime[ID] - Perf_StartCTime[ID];
	return ticks / (CLOCKS_PER_SEC / 1000);
}


//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	perf_c::invalid() const
{
	// Check magic number:
	if(Magic != PERF_MAGIC)
		return "wrong magic number";

	// Check global variables of this module:
	if(Perf_NumUsed < 0)
		return "Perf_NumUsed is negative";

	// Check ID:
	if(ID < -1)
		return "ID is negative (and not -1)";
	if(ID >= VER_MAX_TIMERS)
		return "ID is too large";
	if(ID > Perf_NumUsed)
		return "ID > Perf_NumUsed";
	if(ID != -1 && !Perf_InUse[ID])
		return "Place in Arrays is not reserved";

	// Ok:
	return 0;
}

#endif

